from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

str = None

__metaclass__ = type
__all__ = []

from datetime import datetime
import os
import platform
import pwd
import re
from subprocess import (
    PIPE,
    Popen,
)
from time import (
    sleep,
    time,
)

import apt
from testtools.content import text_content
from testtools.matchers import Contains


def change_logs_permissions(log_dest):
    """Change logs permissions so auto-package-testing can access them."""
    ubuntu_user = pwd.getpwnam('ubuntu')
    uid, gid = ubuntu_user.pw_uid, ubuntu_user.pw_gid
    for root, dirs, files in os.walk(log_dest):
        for d in dirs:
            os.chown(os.path.join(root, d), uid, gid)
        for f in files:
            os.chown(os.path.join(root, f), uid, gid)


def is_juju_core():
    """Report if 'juju' is provider by juju-core."""
    retcode, out, err = run_command(['juju', '--version'])
    # pyjuju 0.6 has no --version option.  pyjuju 0.7 returns "juju
    # <version>".  gojuju returns "<version>".
    # At this point, out should be a string (it's decoded at run_command).
    if out == '' or 'juju ' in out:
        return False
    return True


def is_dist(distname):
    _, _, distid = platform.linux_distribution()
    return distid == distname


def get_ubuntu_version():
    return platform.linux_distribution()[1]


def is_precise():
    return is_dist('precise')


def is_saucy():
    return is_dist('saucy')


def run_command(args, env=None):
    """A wrapper to Popen to run commands in the command-line."""
    process = Popen(args, stdout=PIPE, stderr=PIPE, stdin=PIPE, env=env)
    stdout, stderr = process.communicate()
    return process.returncode, stdout.decode('utf-8'), stderr.decode('utf-8')


def retries(timeout=30, delay=1):
    """Helper for retrying something, sleeping between attempts.

    Yields ``(elapsed, remaining)`` tuples, giving times in seconds.

    @param timeout: From now, how long to keep iterating, in seconds.
    @param delay: The sleep between each iteration, in seconds.
    """
    start = time()
    end = start + timeout
    for now in iter(time, None):
        if now < end:
            yield now - start, end - now
            sleep(min(delay, end - now))
        else:
            break


def add_log_content(runner, log_file):
    """Add content of log_file as detail to the test runner."""
    try:
        log_file_content = open(log_file, 'rb').read().decode('utf-8')
        runner.addDetail(
            "[%s] File %s content" % (datetime.now(), log_file),
            text_content(log_file_content))
    except IOError:
        runner.addDetail(
            "[%s] File %s content" % (datetime.now(), log_file),
            text_content("could not be read"))


def assertStartedUpstartService(runner, service_name, log_file=None):
    """Assert that the upstart service 'service_name' is running."""
    # The service might be running pre-start, retry a few times
    # before giving up.
    expected = service_name + " start/running"
    for elapsed, remaining in retries(timeout=60, delay=5):
        if log_file is not None:
            add_log_content(runner, log_file)
        retcode, output, err = run_command(["service", service_name, "status"])
        if expected in output:
            break

    runner.addDetail(
        "[%s] Last output content" % datetime.now(), text_content(output))
    runner.assertThat(output, Contains(expected))


def assertCommandReturnCode(
        runner, cmd, expected_output, env=None, expected_retcode=0):
    """Assert that cmd returns appropriate return code."""
    if env is None:
        env = os.environ.copy()
    retcode, output, err = run_command(cmd, env=env)
    runner.addDetail('%s stdout' % cmd, text_content(output))
    runner.addDetail('%s stderr' % cmd, text_content(err))
    runner.assertThat(output, Contains(expected_output))
    runner.assertIs(retcode, expected_retcode)


def get_maas_version():
    """Get the version of the installed MAAS package."""
    cache = apt.Cache()
    pkg = cache["maas"]
    return pkg.installed.version


def get_maas_revision():
    """Get the upstream revision of the installed MAAS package.

    This assumes that MAAS is installed."""
    version = get_maas_version()
    # Try to match all the known formats in sequence.
    versions_re = [
        # Daily ppa format (format 1).
        '\d*\+bzr\d*(?:\+dfsg)?-0\+(\d*)\+',
        # Daily ppa format (format 2).
        '\d*\+bzr\d*(?:\+dfsg)?\+(\d*)\+',
        # Release format.
        '\d*\+bzr(\d*)(?:\+dfsg)?',
    ]
    for version_re in versions_re:
        match = re.search(version_re, version)
        if match is not None:
            return int(match.group(1))
    return None


def filter_dict(original, desired_keys):
    """Take a subset of a `dict`.

    :param original: A `dict` (or `OrderedDict`).
    :param desired_keys:  An iterable of keys that should be retained.
        Passing a `dict` (or `OrderedDict`) here will raise an
        `AssertionError`, to make sure that the caller is not passing
        arguments in the wrong order.
    :return: A `dict` that is just like `original` in type and contents,
        except that it only contains those keys that are in `desired_keys`.
        Those entries keep their original values; any other entries from
        `original` will be left out.
    """
    assert isinstance(original, dict), "Original must be a dict."
    assert not isinstance(desired_keys, dict), (
        "Don't pass a dict for desired_keys.  See filter_dict docstring.")
    desired_keys = frozenset(desired_keys)
    return type(original)(
        (key, value)
        for key, value in original.items()
        if key in desired_keys)


def pause_until_released(message):
    """Creates a file `/tmp/paused` and pauses until it is removed so we can
    ssh into the machine and run any sort of test or investigations we need.
    When you are done with your testing, deleting the file will allow the tests
    to continue. The `message` parameter is written to the file.
    """
    pause_file = '/tmp/paused'
    with open(pause_file, 'w') as f:
        f.write("The CI test suite is now PAUSED.\n")
        f.write("To resume, please delete this file.\n")
        f.write("---\n")
        f.write(message)
        f.write("\n")
    while os.path.isfile(pause_file):
        sleep(1)
